from jnt_django_toolbox.forms.widgets.readonly.base import BaseReadOnlyWidget
from jnt_django_toolbox.helpers.object_links import object_change_link


class ForeignKeyReadonlyWidget(BaseReadOnlyWidget):
    """Foreign key readonly widget."""

    def render(self, field_value, field_name, **kwargs) -> str:
        """Render foreign key field."""
        if isinstance(field_value, str):
            return field_value

        return object_change_link(
            field_value,
            field_present=kwargs.get("field_present"),
        )



from typing import Optional

from django.utils.html import format_html

from jnt_django_toolbox.forms.widgets.readonly.base import BaseReadOnlyWidget
from jnt_django_toolbox.helpers.object_links import object_change_link


class GenericForeignKeyReadonlyWidget(BaseReadOnlyWidget):
    """Generic foreign key readonly widget."""

    def render(self, field_value, field_name, **kwargs) -> Optional[str]:
        """Render foreign key field."""
        if not field_value:
            return None

        return format_html(
            '<span class="gfk-object-type">{0}</span> {1}'.format(
                field_value._meta.model_name,
                object_change_link(field_value),
            ),
        )



import jaeger_client
from django.db import connections
from opentracing import global_tracer

from jnt_django_toolbox.profiling.db.jaeger import unwrap_cursor, wrap_cursor
from jnt_django_toolbox.profiling.profilers.base import BaseProfiler


class JaegerProfiler(BaseProfiler):
    """
    Jaeger profiler.

    Spans could be dropped silently, if the size exceeds udp size limit:
    https://github.com/jaegertracing/jaeger-client-node/issues/124#issuecomment-324222456
    Mac users might need to run `sudo sysctl net.inet.udp.maxdgram=65536`
    """

    def __init__(self, service_name: str) -> None:
        """Initializing."""
        self._service_name = service_name
        self._config = jaeger_client.Config(
            config={
                "sampler": {"type": "const", "param": 1},
                "logging": True,
                "max_tag_value_length": 8192,
            },
            service_name=self._service_name,
            validate=True,
        )

    def before_request(self, request, stack) -> None:
        """Start capturing requests."""
        if not self._config.initialized():
            self._config.initialize_tracer()

        stack.enter_context(global_tracer().start_active_span(request.path))
        for connection in connections.all():
            wrap_cursor(connection)

    def after_request(self, request, response):
        """Add profiling info to response."""
        for connection in connections.all():
            unwrap_cursor(connection)



from jnt_django_toolbox.admin.mixins import (
    AutocompleteChangelistFiltersAdminMixin,
    AutocompleteFieldsAdminMixin,
)


class AutocompleteAdminMixin(
    AutocompleteFieldsAdminMixin,
    AutocompleteChangelistFiltersAdminMixin,
):
    """Autocomplete model admin."""



from django.apps import apps
from django.contrib.admin.sites import site
from django.contrib.admin.views.autocomplete import (
    AutocompleteJsonView as BaseAutocompleteJsonView,
)
from django.core.exceptions import PermissionDenied
from django.db import models
from django.http import Http404, JsonResponse


class AutocompleteFieldsError(Exception):
    """Autocomplete fields error."""


class AutocompleteJsonView(BaseAutocompleteJsonView):
    def get(self, request, *args, **kwargs):
        """Override from base."""
        (  # noqa: WPS414
            self.term,
            self.model_admin,
            self.source_field,
            to_field_name,
        ) = self.process_request(request)

        if not self.has_perm(request):
            raise PermissionDenied()

        self._check_inherit_model_admin()

        self.object_list = self.get_queryset()
        context = self.get_context_data()

        return JsonResponse(
            {
                "results": [
                    self.model_admin.autocomplete_item_data(instance, request)
                    for instance in context["object_list"]
                ],
                "pagination": {"more": context["page_obj"].has_next()},
            }
        )

    def get_queryset(self):
        """Additional filter for queryset in autocomplete."""
        queryset = super().get_queryset()

        return self.model_admin.autocomplete_queryset(self.request, queryset)

    def process_request(self, request):  # noqa: WPS238
        """Process request."""
        self._apply_pagination(request)
        try:
            app_label = request.GET["app_label"]
            model_name = request.GET["model_name"]
            field_name = request.GET.get("field_name")
        except KeyError as err:
            raise PermissionDenied from err

        if field_name:
            return super().process_request(request)

        try:
            source_model = apps.get_model(app_label, model_name)
        except LookupError as err:
            raise PermissionDenied from err

        try:
            model_admin = self.admin_site._registry[source_model]
        except KeyError as err:
            raise PermissionDenied from err

        if not model_admin.get_search_fields(request):
            raise Http404(
                "{0} must have search_fields for the autocomplete_view.".format(
                    type(model_admin).__qualname__,
                )
            )

        to_field_name = "id"
        source_field = models.ForeignKey(
            source_model,
            on_delete=models.CASCADE,
        )

        return (
            request.GET.get("term", ""),
            model_admin,
            source_field,
            to_field_name,
        )

    def _apply_pagination(self, request) -> None:
        page_size = request.GET.get("page_size")
        if page_size:
            self.paginate_by = int(page_size)

    def _check_inherit_model_admin(self) -> None:
        """Check inherit model_admin."""
        from jnt_django_toolbox.admin.mixins import (
            AutocompleteFieldsAdminMixin,
        )

        if not isinstance(self.model_admin, AutocompleteFieldsAdminMixin):
            raise AutocompleteFieldsError(
                "{0} is necessary to inherit the class from the AutocompleteAdminMixin".format(
                    self.model_admin,
                ),
            )


def autocomplete_view(request):
    return AutocompleteJsonView.as_view(admin_site=site)(request)



from django.contrib.admin.helpers import AdminReadonlyField
from django.http import HttpRequest

from jnt_django_toolbox.admin.fields import AdminChangeListField
from jnt_django_toolbox.admin.mixins import ReadonlyWidgetsMixin
from jnt_django_toolbox.helpers.readonly_form_field import (
    get_present_admin_readonly_field,
)

# Patching AdminReadonlyField
orig_contents = AdminReadonlyField.contents


def contents(self):
    return get_present_admin_readonly_field(self) or orig_contents(self)


AdminReadonlyField.contents = contents


class ClickableLinksAdminMixin(ReadonlyWidgetsMixin):
    def get_list_display(self, request: HttpRequest, *args, **kwargs):
        list_display = super().get_list_display(request, *args, **kwargs)

        if len(list_display) < 2:
            return list_display

        updated_list = [list_display[0]]
        updated_list.extend(
            [
                self._update_column_field(column_field, request)
                for column_field in list_display[1:]
            ],
        )

        return tuple(updated_list)

    def _update_column_field(self, column_field, request=None):
        """Update column field."""
        field_present = None
        field_name = column_field
        if "__" in field_name:
            field_name, field_present = field_name.split("__")

        readonly_widget = self.readonly_widget(field_name, request)
        if not readonly_widget:
            return column_field

        return AdminChangeListField(
            model_admin=self,
            field_name=field_name,
            readonly_widget=readonly_widget,
            field_present=field_present,
        )



import typing as ty

from django.contrib import admin
from django.contrib.admin.sites import site

from jnt_django_toolbox.admin.views.content_type_autocomplete import (
    ContentTypeAutocompleteView,
)


def get_model_admin(model) -> ty.Optional[admin.ModelAdmin]:
    return site._registry.get(model)


class BaseContentTypeAdmin(admin.ModelAdmin):
    """Base content type admin."""

    search_fields = ("model",)

    def has_add_permission(self, request):
        """Check has add permissions."""
        return False

    def has_change_permission(self, request, obj=None):
        """Check has change permissions."""
        return False

    def has_view_or_change_permission(self, request, obj=None):
        """Check has view or change permissions."""
        return False

    def has_view_permission(self, request, obj=None):
        """Check has view permissions."""
        return True

    def get_model_perms(self, request):
        """Return a dict of all perms for this model."""
        return {}

    def autocomplete_view(self, request):
        """Content type autocomplete view."""
        return ContentTypeAutocompleteView.as_view(model_admin=self)(request)

    def autocomplete_queryset(self, request, queryset):
        queryset = super().autocomplete_queryset(request, queryset)

        ids = []
        for content_type in queryset:
            model_class = content_type.model_class()

            if get_model_admin(model_class):
                ids.append(content_type.pk)

        return queryset.filter(id__in=ids)

    def autocomplete_item_data(self, instance, request):
        item_data = super().autocomplete_item_data(instance, request)

        model = instance.model_class()

        related_obj = None

        for related_object in model._meta.related_objects:
            model_admin = get_model_admin(related_object.related_model)
            if model_admin:
                related_obj = related_object
                break

        if not related_obj:
            return item_data

        item_data.update(
            {
                "app_label": related_obj.related_model._meta.app_label,
                "model": related_obj.related_model._meta.model_name,
                "field_name": related_obj.remote_field.name,
                "field_placeholder": model._meta.verbose_name.title(),
            }
        )

        return item_data



from django.contrib import admin

from jnt_django_toolbox.admin.mixins import (
    AutocompleteAdminMixin,
    ClickableLinksAdminMixin,
)


class BaseModelAdmin(
    ClickableLinksAdminMixin,
    AutocompleteAdminMixin,
    admin.ModelAdmin,
):
    """Base model admin."""



from .autocomplete_fields import AutocompleteFieldsAdminMixin
from .generic_foreign_key import GenericForeignKeyAdminMixin
from .generic_foreign_key_inline import GenericForeignKeyInlineAdminMixin
from .readonly_widgets import ReadonlyWidgetsMixin
from .clickable_links import ClickableLinksAdminMixin
from .autocomplete_changelist_filters import (
    AutocompleteChangelistFiltersAdminMixin,
)
from .autocomplete import AutocompleteAdminMixin



from contextlib import contextmanager
from time import monotonic

from django.core.cache import cache

LOCK_EXPIRE = 60 * 10  # Lock expires in 10 minutes


@contextmanager
def global_lock(lock_id, value=1, expire=LOCK_EXPIRE):
    """
    Global lock mechanizm.

    Idea from
    http://docs.celeryproject.org/en/latest/tutorials/task-cookbook.html
    """
    timeout_at = monotonic() + expire - 3
    lock_key = build_global_cache_key(lock_id)
    # cache.add fails if the key already exists
    status = cache.add(lock_key, value, expire)
    try:  # noqa: WPS501
        yield status
    finally:
        # memcache delete is very slow, but we have to use it to take
        # advantage of using add() for atomic locking
        if monotonic() < timeout_at and status:
            # don't release the lock if we exceeded the timeout
            # to lessen the chance of releasing an expired lock
            # owned by someone else.
            cache.delete(lock_key)


def build_global_cache_key(lock_id):
    """Function for build global lock key."""
    return "____{0}____".format(lock_id)



from urllib.parse import urlencode

from django.contrib import admin
from django.urls import NoReverseMatch, reverse


class _AdminUrlProvider:
    allowed_urls_types = (
        "add",
        "change",
        "delete",
        "changelist",
        "view",
        "autocomplete",
    )

    def for_view(
        self,
        url_type: str,
        model,
        model_admin: admin.ModelAdmin | None = None,
    ) -> str | None:
        if url_type not in self.allowed_urls_types:
            raise "Type not allowed, allowed: {0}".format(
                self.allowed_urls_types,
            )

        model_id = getattr(model, "id", None)
        view_name = self._make_view_name(url_type, model, model_admin)
        return self._reverse(
            view_name,
            args=[model_id]
            if model_id and url_type not in {"add", "changelist"}
            else None,
        )

    def add_url(self, model):
        return self.for_view("add", model)

    def change_url(
        self,
        model,
        model_admin: admin.ModelAdmin | None = None,
    ) -> str | None:
        return self.for_view("change", model, model_admin)

    def delete_url(
        self,
        model,
        model_admin: admin.ModelAdmin | None = None,
    ) -> str | None:
        return self.for_view("delete", model, model_admin)

    def view_url(
        self,
        model,
        model_admin: admin.ModelAdmin | None = None,
    ) -> str | None:
        return self.for_view("view", model, model_admin)

    def list_url(
        self,
        model,
        model_admin: admin.ModelAdmin | None = None,
    ) -> str | None:
        return self.for_view("changelist", model, model_admin)

    def autocomplete_url(
        self,
        model,
        model_admin: admin.ModelAdmin | None = None,
    ) -> str | None:
        site = self._get_admin_site(model_admin)

        query_params = {
            "app_label": model._meta.app_label,
            "model_name": model._meta.model_name,
            "field_name": "",
        }
        related_objects = model._meta.related_objects
        if related_objects:
            remote_field = related_objects[0].remote_field
            query_params.update(
                {
                    "app_label": remote_field.model._meta.app_label,
                    "model_name": remote_field.model._meta.model_name,
                    "field_name": remote_field.name,
                },
            )

        return "{0}?{1}".format(
            reverse("{0}:autocomplete".format(site.name)),
            urlencode(query_params),
        )

    def __call__(self, model):
        return {
            url_type: self.for_view(url_type, model)
            for url_type in self.allowed_urls_types
        }

    def _make_view_name(
        self,
        url_type: str,
        model,
        model_admin: admin.ModelAdmin | None = None,
    ) -> str:
        site = self._get_admin_site(model_admin)

        return "{0}:{1}_{2}_{3}".format(
            site.name,
            model._meta.app_label,
            model._meta.model_name,
            url_type,
        )

    def _reverse(self, view_name: str, **kwargs) -> str | None:
        try:
            return reverse(view_name, **kwargs)
        except NoReverseMatch:
            return None

    def _get_admin_site(
        self,
        model_admin: admin.ModelAdmin | None = None,
    ) -> admin.AdminSite:
        if model_admin:
            return model_admin.admin_site

        return admin.site


admin_url_provider = _AdminUrlProvider()



from http import HTTPStatus

from django.contrib.admin.views.autocomplete import AutocompleteJsonView
from django.http import Http404, JsonResponse
from django.urls import reverse
from django.utils.text import capfirst

from jnt_django_toolbox.admin.helpers.urls import admin_url_provider


class ContentTypeAutocompleteView(AutocompleteJsonView):
    # copied from base class
    def get(self, request, *args, **kwargs):
        if not self.model_admin.get_search_fields(request):
            msg_template = (
                "{0} must have search_fields for the autocomplete_view."
            )
            raise Http404(msg_template.format(type(self.model_admin).__name__))
        if not self.has_perm(request):
            return JsonResponse(
                {"error": "403 Forbidden"},
                status=HTTPStatus.FORBIDDEN,
            )

        self.term = request.GET.get("term", "")
        self.paginator_class = self.model_admin.paginator
        self.object_list = self.filter_queryset(self.get_queryset())
        context = self.get_context_data()

        return JsonResponse(
            {
                "results": [
                    {
                        "id": str(content_type.pk),
                        "text": capfirst(content_type.model),
                        # modify
                        "autocomplete_url": self.get_autocomplete_url(
                            content_type
                        ),
                        "data_app": content_type.app_label,
                        "data_model": content_type.model,
                        "data_change_url": self._get_change_url_template(
                            content_type
                        ),
                    }
                    for content_type in context["object_list"]
                ],
                "pagination": {"more": context["page_obj"].has_next()},
            }
        )

    def get_autocomplete_url(self, content_type):
        return admin_url_provider.autocomplete(content_type.model_class())

    def filter_queryset(self, queryset):
        return self._filter_queryset_by_ids(queryset)

    def _filter_queryset_by_ids(self, queryset):
        ids = self.request.GET.getlist("ids[]")
        if ids:
            queryset = queryset.filter(id__in=ids)
        return queryset

    def _get_change_url_template(self, content_type):
        list_url = reverse(
            "admin:{0}_{1}_changelist".format(
                content_type.app_label,
                content_type.model,
            ),
        )
        return "{0}{{id}}/change/".format(list_url)



from jnt_django_toolbox.forms.widgets.readonly.base import BaseReadOnlyWidget
from jnt_django_toolbox.helpers.object_links import get_display_for_many


class ManyToManyReadonlyWidget(BaseReadOnlyWidget):
    """Many to many readonly widget."""

    def render(self, field_value, field_name, **kwargs) -> str:
        """Render m2m field."""
        if isinstance(field_value, str):
            return field_value

        return get_display_for_many(
            field_value.all(),
            field_present=kwargs.get("field_present"),
        )



import types
import typing as ty

from django.contrib.auth.models import Permission
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.http import HttpRequest

from jnt_django_toolbox.forms.widgets.readonly import (
    ForeignKeyReadonlyWidget,
    GenericForeignKeyReadonlyWidget,
    ManyToManyReadonlyWidget,
    PermissionSelectMultipleReadonlyWidget,
    StringReadonlyWidget,
)
from jnt_django_toolbox.forms.widgets.readonly.base import BaseReadOnlyWidget
from jnt_django_toolbox.models.fields import GenericForeignKey

READONLY_WIDGETS = types.MappingProxyType(
    {
        models.ForeignKey: ForeignKeyReadonlyWidget(),
        models.OneToOneField: ForeignKeyReadonlyWidget(),
        models.ManyToManyField: ManyToManyReadonlyWidget(),
        models.ManyToOneRel: ManyToManyReadonlyWidget(),
        models.CharField: StringReadonlyWidget(),
        models.TextField: StringReadonlyWidget(),
        GenericForeignKey: GenericForeignKeyReadonlyWidget(),
    },
)


class ReadonlyWidgetsMixin:
    def readonly_widget(
        self,
        field_name: str,
        request: HttpRequest = None,
    ) -> ty.Optional[BaseReadOnlyWidget]:
        """Return readonly widget for admin readonly field."""
        if callable(field_name) or getattr(self, field_name, None):
            return None

        try:
            model_field = self.model._meta.get_field(field_name)
        except FieldDoesNotExist:
            return None

        field_class = model_field.__class__

        if issubclass(field_class, models.ManyToManyField) and issubclass(
            model_field.related_model,
            Permission,
        ):
            return PermissionSelectMultipleReadonlyWidget()

        return self.readonly_widgets(request).get(field_class)

    def readonly_widgets(
        self,
        request: HttpRequest = None,
    ) -> ty.Dict[models.Field, BaseReadOnlyWidget]:
        """Return available readonly_widgets."""
        return dict(READONLY_WIDGETS)
